import abc
from typing import Dict, Any, Type, TypeVar

from exco.cell_full_path import CellFullPath
from exco.cell_location import CellLocation
from exco.extractor.actor import Actor
from exco.extractor.locator.locating_result import LocatingResult
from openpyxl import Workbook

T = TypeVar('T')


class Locator(Actor, abc.ABC):
    @abc.abstractmethod
    def locate(self, anchor_cell_location: CellLocation, workbook: Workbook) -> LocatingResult:
        raise NotImplementedError()



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import List, Iterable, Dict, Any

import yaml
from exco import setting
from exco.exception import TooManyBeginException, TooManyEndException, BadTemplateException, ExcoException, \
    TableKeyNotFound, ExpectEndException, YamlParseError
from exco.extractor_spec import CellExtractionSpec
from exco.extractor_spec.spec_source import SpecSource
from exco.setting import k_table_key
from yaml.scanner import ScannerError


@dataclass
class LineCollector:
    start_line: int
    raw: List[str] = field(default_factory=list)
    end_line: int = 0

    def __post_init__(self):
        self.end_line = self.start_line

    def append(self, i: int, line: str):
        self.end_line = i
        self.raw.append(line)


@dataclass
class ExcoBlock(SpecSource):
    start_line: int
    end_line: int
    raw: str

    def table_key(self) -> str:
        """

        Returns:
            str. table key for this block.
        """
        d = self.to_dict()
        try:
            return d[k_table_key]
        except LookupError as e:
            raise TableKeyNotFound() from e

    @property
    def key(self) -> str:
        """

        Returns:
            str. key for this block
        """
        return self.to_dict()[setting.k_key]

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary.

        Returns:
            Dict[str, Any]
        """
        try:
            return yaml.load(self.raw, Loader=yaml.FullLoader)
        except ScannerError as e:
            raise YamlParseError(self.raw) from e

    def to_cell_extractor_task_spec(self) -> CellExtractionSpec:
        """To CellExtractionSpec

        Returns:
            CellExtractionSpecs
        """
        return CellExtractionSpec.from_dict(self.to_dict(), source=self)

    def describe(self) -> str:
        """

        Returns:
            String description of this block.
        """
        return self.raw

    @classmethod
    def from_line_collector(cls, line_collector: LineCollector) -> 'ExcoBlock':
        """Construct from line_collector.

        Args:
            line_collector (LineCollector):

        Returns:
            ExcoBlock
        """
        return ExcoBlock(
            start_line=line_collector.start_line,
            end_line=line_collector.end_line,
            raw='\n'.join(line_collector.raw)
        )

    @classmethod
    def simple(cls, raw: str, start_line: int = 0, end_line: int = 0) -> 'ExcoBlock':
        return ExcoBlock(
            raw=raw,
            start_line=start_line,
            end_line=end_line
        )


@dataclass
class ExcoBlockCollection:
    table_blocks: List[ExcoBlock]
    cell_blocks: List[ExcoBlock]
    column_blocks: List[ExcoBlock]

    def n_total_blocks(self) -> int:
        """

        Returns:
            total number of blocks.
        """
        return len(self.table_blocks) + len(self.cell_blocks) + len(self.column_blocks)

    @classmethod
    def from_string(cls, block: str):
        """Construct from comment string."""
        return ExcoBlockParser.parse(block.split('\n'))


class ExcoBlockParserState(Enum):
    OUTSIDE = auto()
    INSIDE_CELL = auto()
    INSIDE_TABLE = auto()
    INSIDE_COL = auto()

    def is_inside(self):
        return self != self.OUTSIDE

    def is_outside(self):
        return self == self.OUTSIDE


S = ExcoBlockParserState


class ExcoBlockParser:
    """
    FSM Parser for comment block.
    """

    @classmethod
    def enter_block(cls, cs: S, line: str, ns: S) -> S:
        """Enter block transition

        Args:
            cs (ExcoBlockParserState):
            line (str): this line
            ns (ExcoBlockParserState):

        Raises:
            TooManyBeginException. If try to enter when it's already enter.
        Returns:
            ExcoBlockParserState. Next State.
        """
        if cs.is_outside():
            return ns
        else:
            raise TooManyBeginException(f'Too many begin exception for {line}')

    @classmethod
    def compute_next_state(cls, current_state: S, line: str) -> S:
        """Compute the next state

        Args:
            current_state (ExcoBlockParserState):
            line (str):

        Raises:
            ExcoException on invalid transition

        Returns:
            ExcoBlockParserState. Next state.
        """
        cs = current_state
        if line == setting.start_table_marker:
            return cls.enter_block(cs, line, S.INSIDE_TABLE)
        elif line == setting.start_cell_marker:
            return cls.enter_block(cs, line, S.INSIDE_CELL)
        elif line == setting.start_col_marker:
            return cls.enter_block(cs, line, S.INSIDE_COL)
        elif line == setting.end_marker:
            if current_state.is_inside():
                return S.OUTSIDE
            else:
                raise TooManyEndException(f'Too many end markers.')
        else:
            return current_state

    @classmethod
    def parse(cls, lines: Iterable[str]) -> ExcoBlockCollection:
        """Parse the line iterable into ExcoBlockCollection

        Args:
            lines (Iterable[str]):

        Returns:
            ExcoBlockCollection
        """

        cs = S.OUTSIDE
        collectors: Dict[S, List[LineCollector]] = {
            S.INSIDE_TABLE: [],
            S.INSIDE_CELL: [],
            S.INSIDE_COL: []
        }
        current_collector = None
        try:
            for i, line in enumerate(lines, start=1):
                ns = cls.compute_next_state(cs, line)
                if cs.is_outside() and ns.is_inside():  # enter block
                    current_collector = LineCollector(i)
                elif cs.is_inside() and ns.is_inside():  # stay inside -> collect
                    current_collector.append(i, line)
                elif cs.is_inside() and ns.is_outside():  # exit block -> put in collectors
                    collectors[cs].append(current_collector)
                    current_collector = None
                cs = ns
            if cs.is_inside():
                raise ExpectEndException()

            return ExcoBlockCollection(
                table_blocks=[ExcoBlock.from_line_collector(lc) for lc in collectors[S.INSIDE_TABLE]],
                cell_blocks=[ExcoBlock.from_line_collector(lc) for lc in collectors[S.INSIDE_CELL]],
                column_blocks=[ExcoBlock.from_line_collector(lc) for lc in collectors[S.INSIDE_COL]]
            )
        except ExcoException as e:
            raise BadTemplateException(f'Bad template at Line {i}\n{lines}') from e



from dataclasses import dataclass

from exco.extractor.table_end_conditions.table_end_condition_param import TableEndConditionParam
from exco.extractor.table_end_conditions.table_end_condition_result import TableEndConditionResult
from exco.extractor.table_end_conditions.table_end_condition import TableEndCondition


@dataclass
class AllBlankTableEndCondition(TableEndCondition):
    """
    End Condition when all cell in the row is blank.
    """

    def test(self, param: TableEndConditionParam) -> TableEndConditionResult:
        """ Return True if all cells in this table row is blank.

        Args:
            param (TableEndConditionParam):

        Returns:
            TableEndConditionResult
        """
        all_blanks = all(cfp.is_blank() for key, cfp in param.cfps.items())
        return TableEndConditionResult.good(
            should_terminate=all_blanks,
            is_inclusive=False
        )



from dataclasses import dataclass
from typing import Optional


@dataclass
class TableEndConditionResult:
    """
    Table End Conditon Result
    """
    should_terminate: bool
    is_inclusive: bool  # should the row that end condition evaluates to true be included in the results
    is_ok: bool
    msg: str = ''
    exception: Optional[Exception] = None

    @property
    def is_exclusive(self) -> bool:
        """

        Returns:
            bool. True if the row that the condition is evaluated to True should be excluded
            from the result.
        """
        return not self.is_inclusive

    @classmethod
    def good(cls, should_terminate: bool, is_inclusive: bool) -> 'TableEndConditionResult':
        """Good TableEndConditionResult

        Args:
            should_terminate (bool):
            is_inclusive (bool):

        Returns:
            TableEndConditionResult
        """
        return TableEndConditionResult(
            should_terminate=should_terminate,
            is_inclusive=is_inclusive,
            is_ok=True
        )

    @classmethod
    def bad(cls, msg: str = '', exception: Optional[Exception] = None) -> 'TableEndConditionResult':
        """Bad TableEndConditionResult

        Args:
            msg (str):
            exception ():

        Returns:
            TableEndConditionResult
        """
        return TableEndConditionResult(
            should_terminate=True,
            is_inclusive=False,
            is_ok=False,
            msg=msg,
            exception=exception
        )



from dataclasses import dataclass, field
from typing import Dict, List, Any

from exco import CellLocation
from exco.exception import TooManyRowRead, NoEndConditonError
from exco.extractor.locator.locator import Locator
from exco.extractor.cell_extraction_task import CellExtractionTaskResult, CellExtractionTask
from exco.extractor.locator.locating_result import LocatingResult
from exco.extractor.table_end_conditions.built_in.all_blank_table_end_condition import AllBlankTableEndCondition
from exco.extractor.table_end_conditions.table_end_condition import TableEndCondition
from exco.extractor.table_end_conditions.table_end_condition_factory import TableEndConditionFactory
from exco.extractor.table_end_conditions.table_end_condition_param import TableEndConditionParam
from exco.extractor.table_end_conditions.table_end_condition_result import TableEndConditionResult
from exco.extractor_spec.table_extraction_spec import TableItemDirection, TableEndConditionSpec
from openpyxl import Workbook
from exco import setting as st


@dataclass
class EndConditionCollectionResult:
    end_condition_results: List[TableEndConditionResult]

    @property
    def is_ok(self) -> bool:
        """

        Returns:
            bool. True if the result is ok.
        """
        return all(x.is_ok for x in self.end_condition_results)

    def should_terminate_exclusively(self) -> bool:
        """

        Returns:
            bool. should terminate WITHOUT parsing this row
        """
        return any(ecr.should_terminate and ecr.is_exclusive for ecr in self.end_condition_results)

    def should_terminate_inclusively(self) -> bool:
        """

        Returns:
            bool. should terminate AFTER parsing this row
        """
        return any(ecr.should_terminate and ecr.is_inclusive for ecr in self.end_condition_results)


@dataclass
class EndConditionCollection:
    """Collection of End Condtition.
    When it's tested all the endcondition are applied sequentially.
    """
    end_conditions: List[TableEndCondition]

    def test(self, param: TableEndConditionParam) -> EndConditionCollectionResult:
        return EndConditionCollectionResult(
            end_condition_results=[ec.test(param) for ec in self.end_conditions]
        )

    @classmethod
    def from_spec(cls,
                  specs: List[TableEndConditionSpec],
                  factory: TableEndConditionFactory) -> 'EndConditionCollection':
        if not specs:
            raise NoEndConditonError()
        ecs = [factory.create_from_spec(spec) for spec in specs]

        return EndConditionCollection(end_conditions=ecs)

    @classmethod
    def default(cls) -> 'EndConditionCollection':
        return EndConditionCollection([AllBlankTableEndCondition()])


@dataclass
class RowExtractionTaskResult:
    """Extraction Result for a Row"""
    cell_results: Dict[str, CellExtractionTaskResult]

    @property
    def is_ok(self) -> bool:
        return all(x.is_ok for x in self.cell_results.values())

    def to_dict(self) -> Dict[str, Any]:
        return {k: v.get_value(None) for k, v in self.cell_results.items()}


@dataclass
class TableExtractionTaskResult:
    """Extraction Result for a Table"""
    key: str
    locating_result: LocatingResult
    row_results: List[RowExtractionTaskResult] = field(default_factory=list)
    end_condition_results: List[EndConditionCollectionResult] = field(default_factory=list)

    @property
    def is_ok(self):
        return self.locating_result.is_ok and \
               all(rr.is_ok for rr in self.row_results) and \
               all(ec.is_ok for ec in self.end_condition_results)

    @classmethod
    def fail_locating_result(cls, key: str, locating_result: LocatingResult) -> 'TableExtractionTaskResult':
        """TableExtractionTaskResult when it fails locating the cell

        Args:
            locating_result (LocatingResult):

        Returns:
            TableExtractionTaskResult
        """
        return TableExtractionTaskResult(key=key, locating_result=locating_result)

    def get_value(self) -> List[Dict[str, Any]]:
        """
        Returns:
            Python Equivalent Value
        """
        return [res.to_dict() for res in self.row_results]


@dataclass
class TableExtractionTask:
    key: str
    locator: Locator
    columns: Dict[int, CellExtractionTask]  # offset -> APVSpec
    end_condition: EndConditionCollection  # there is inclusive and exclusive
    item_direction: TableItemDirection = TableItemDirection.DOWNWARD

    def shift_item_direction(self, cl: CellLocation, offset: int = 1) -> CellLocation:
        """Shift cell location in the item direction.

        Args:
            cl (CellLocation):
            offset (int):

        Returns:
            CellLocation
        """
        if self.item_direction == TableItemDirection.DOWNWARD:
            return cl.shift_row(offset)
        else:
            return cl.shift_col(offset)

    def shift_column_direction(self, cl: CellLocation, offset: int) -> CellLocation:
        """Shift cell location in the column direction.

        Args:
            cl ():
            offset ():

        Returns:

        """
        if self.item_direction == TableItemDirection.DOWNWARD:
            return cl.shift_col(offset)
        else:
            return cl.shift_row(offset)

    def _build_row_cell_locations(self, key_cell: CellLocation) -> Dict[str, CellLocation]:
        """Build a dictionary from column key to each columns' CellLocation.
        The offset is applied and the key is change to column key

        Args:
            key_cell (CellLocation):

        Returns:
            Dict[str, CellLocation]. column key -> CellLocation
        """
        return {cet.key: self.shift_column_direction(key_cell, offset) for offset, cet in self.columns.items()}

    def process(self, anchor_cell_location: CellLocation, workbook: Workbook) -> TableExtractionTaskResult:
        """ Extract table from work book as if the anchor location is at anchor_cell_location

        Args:
            anchor_cell_location (CellLocation):
            workbook (Workbook):

        Returns:

        """
        locating_result = self.locator.locate(anchor_cell_location, workbook)
        if not locating_result.is_ok:
            return TableExtractionTaskResult.fail_locating_result(key=self.key, locating_result=locating_result)

        irow = 0
        should_terminate = False
        key_cell = anchor_cell_location
        row_results = []
        end_condition_results = []
        while not should_terminate:
            irow += 1
            if irow >= st.table_infinite_loop_guard:
                raise TooManyRowRead(f'setting.table_infinite_loop_guard ({st.table_infinite_loop_guard}) reached')
            cell_locations = self._build_row_cell_locations(key_cell)
            cfps = {k: cl.get_cell_full_path(workbook) for k, cl in cell_locations.items()}

            # Test end condition
            ec_param = TableEndConditionParam(row_count=irow, cfps=cfps)
            ec_results = self.end_condition.test(ec_param)
            end_condition_results.append(ec_results)
            if ec_results.should_terminate_exclusively():
                break

            # parse
            cell_results = []
            for offset, cet in self.columns.items():
                cell_cl = self.shift_column_direction(key_cell, offset)
                cell_results.append(cet.process(cell_cl, workbook))
            row_results.append(RowExtractionTaskResult(
                {cr.key: cr for cr in cell_results}
            ))
            if ec_results.should_terminate_inclusively():
                break

            # shift
            key_cell = self.shift_item_direction(key_cell)

        return TableExtractionTaskResult(
            key=self.key,
            locating_result=locating_result,
            row_results=row_results,
            end_condition_results=end_condition_results
        )



from dataclasses import dataclass, field
from typing import Dict, Any, ClassVar, List, Set

from exco.extractor_spec.validator_spec import ValidatorSpec
from exco.extractor_spec.assumption_spec import AssumptionSpec
from exco.extractor_spec.parser_spec import ParserSpec
from exco.extractor_spec.spec_source import SpecSource, UnknownSource
from exco import setting as st


@dataclass
class APVSpec:  # Assume Parse Validate
    key: str
    parser: ParserSpec
    validations: Dict[str, ValidatorSpec] = field(default_factory=dict)
    assumptions: Dict[str, AssumptionSpec] = field(default_factory=dict)
    source: SpecSource = field(default_factory=UnknownSource)

    consumed_keys: ClassVar[Set[str]] = {st.k_key, st.k_validations, st.k_assumptions}
    allowed_keys: ClassVar[Set[str]] = consumed_keys | ParserSpec.allowed_keys

    @classmethod
    def from_dict(cls, d: Dict[str, Any], source: source) -> 'APVSpec':
        return APVSpec(
            key=d[st.k_key],
            parser=ParserSpec.from_dict(d),
            validations={v[st.k_key]: ValidatorSpec.from_dict(v) for v in d.get(st.k_validations, [])},
            assumptions={v[st.k_key]: AssumptionSpec.from_dict(v) for v in d.get(st.k_assumptions, [])},
            source=source if source is not None else UnknownSource()
        )




start_cell_marker = '{{--'
start_table_marker = '{{--table'
start_col_marker = '{{--col'
end_marker = '--}}'

default_locator = 'at_comment_cell'

# keys:

k_key = 'key'
k_name = 'name'
k_parser = 'parser'
k_params = 'params'
k_locator = 'locator'
k_validations = 'validations'
k_assumptions = 'assumptions'

k_table_key = 'table_key'
k_columns = 'columns'
k_end_conditions = 'end_conditions'
k_item_direction = 'item_direction'

table_infinite_loop_guard = 10000



from dataclasses import dataclass, field
from typing import Dict, Any

from exco.extractor_spec.type import SpecParam
from exco.util import name_params


@dataclass
class AssumptionSpec:
    """
    Assumption is something to check before parsing
    Ex: if the column on the left has the correct label
    """
    name: str
    params: SpecParam = field(default_factory=dict)

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> 'AssumptionSpec':
        """Construct AssumptionTaskSpec

        Args:
            d ():

        Returns:

        """
        name, params = name_params(d)
        return AssumptionSpec(name=name, params=params)



import abc
from dataclasses import dataclass, field
from typing import Generic, Dict, TypeVar

from exco.cell_location import CellLocation
from exco.extractor.assumption.assumption import Assumption
from exco.extractor.assumption.assumption_result import AssumptionResult
from exco.extractor.locator.built_in.at_comment_cell_locator import AtCommentCellLocator
from exco.extractor.locator.locator import Locator
from exco.extractor.locator.locating_result import LocatingResult
from exco.extractor.parser.parser import Parser

from exco.extractor.parser.parsing_result import ParsingResult
from exco.extractor.validator.validation_result import ValidationResult
from exco.extractor.validator.validator import Validator
from exco.util import long_string
from openpyxl import Workbook

T = TypeVar('T')


@dataclass
class CellExtractionTaskResult(Generic[T]):
    key: str
    locating_result: LocatingResult
    parsing_result: ParsingResult[T]
    validation_results: Dict[str, ValidationResult] = field(default_factory=dict)
    assumption_results: Dict[str, AssumptionResult] = field(default_factory=dict)

    def get_value(self, default: T) -> T:
        """Get Python equivalent value

        Args:
            default (T):

        Returns:
            T. Python equivalent value.
        """
        return self.parsing_result.get_value(default=default)

    def is_ok(self):
        """
        Returns:
            True if it pass assumption, parsing, and validation.
        """
        return all(ar.is_ok for ar in self.assumption_results.values()) and \
               self.parsing_result.is_ok and \
               all(vr.is_ok for vr in self.validation_results.values())

    @classmethod
    def fail_locating(cls, key: str, locating_result: LocatingResult) -> 'CellExtractionTaskResult':
        msg = "Fail Locating"
        assert not locating_result.is_ok
        return CellExtractionTaskResult(
            key=key,
            locating_result=locating_result,
            parsing_result=ParsingResult.bad(msg=msg)
        )

    @classmethod
    def fail_assumptions(cls, key: str, locating_result: LocatingResult,
                         assumption_results: Dict[str, AssumptionResult]):
        msg = "Fail Assumption"
        assert any(not ar.is_ok for ar in assumption_results.values())
        return CellExtractionTaskResult(
            key=key,
            locating_result=locating_result,
            assumption_results=assumption_results,
            parsing_result=ParsingResult.bad(msg=msg)
        )


@dataclass
class CellExtractionTask(Generic[T]):
    key: str
    locator: Locator
    parser: Parser[T]
    validators: Dict[str, Validator[T]]
    assumptions: Dict[str, Assumption]

    def __str__(self):
        s = long_string(f"""
        key: "{self.key}"
        locator: {self.locator}
        parser: {self.parser}
        validators: {[dict(key=key, v=v) for key, v in self.validators.items()]}
        assumptions: {[dict(key=key, a=a) for key, a in self.assumptions.items()]}""")
        return s

    def process(self, anchor_cell_location: CellLocation, workbook: Workbook) -> CellExtractionTaskResult:
        locating_result = self.locator.locate(
            anchor_cell_location=anchor_cell_location,
            workbook=workbook
        )
        if not locating_result.is_ok:
            return CellExtractionTaskResult.fail_locating(key=self.key, locating_result=locating_result)
        cfp = locating_result.location.get_cell_full_path(workbook)

        assumption_results = {k: assumption.assume(cfp) for k, assumption in self.assumptions.items()}
        if any(not ar.is_ok for ar in assumption_results.values()):
            return CellExtractionTaskResult.fail_assumptions(
                key=self.key,
                locating_result=locating_result,
                assumption_results=assumption_results
            )

        parsing_result = self.parser.parse(cfp)
        if not parsing_result.is_ok:
            return CellExtractionTaskResult(
                key=self.key,
                locating_result=locating_result,
                assumption_results=assumption_results,
                parsing_result=parsing_result
            )

        validation_results = {k: vt.validate(parsing_result.value) for k, vt in self.validators.items()}
        return CellExtractionTaskResult(
            key=self.key,
            locating_result=locating_result,
            assumption_results=assumption_results,
            parsing_result=parsing_result,
            validation_results=validation_results
        )

    @classmethod
    def simple(cls, key: str, parser: Parser):
        return cls(
            key=key,
            locator=AtCommentCellLocator(),
            parser=parser,
            validators={},
            assumptions={}
        )



from dataclasses import dataclass
from typing import Dict, Any

from exco.cell_full_path import CellFullPath
from exco.extractor.assumption.assumption import Assumption
from exco.extractor.assumption.assumption_result import AssumptionResult


@dataclass
class LeftCellMatchAssumption(Assumption):
    label: str

    def assume(self, cfp: CellFullPath) -> AssumptionResult:
        cond = cfp.shift(col=-1).cell.value == self.label
        if cond:
            return AssumptionResult.good()
        else:
            return AssumptionResult.bad(f'Cell to the left of {cfp.cell.coordinate} does not match {self.label}')



from dataclasses import dataclass
from typing import Optional

from exco import CellLocation


@dataclass
class LocatingResult:
    location: Optional[CellLocation]
    is_ok: bool
    msg: str = ""
    exception: Optional[Exception] = None

    @classmethod
    def good(cls, location: CellLocation) -> 'LocatingResult':
        return LocatingResult(location=location, is_ok=True)

    @classmethod
    def bad(cls, msg: str, exception: Optional[Exception] = None) -> 'LocatingResult':
        return LocatingResult(location=None, is_ok=False, msg=msg, exception=exception)



from dataclasses import dataclass
from typing import Generic, Optional, TypeVar

T = TypeVar('T')


@dataclass
class ParsingResult(Generic[T]):
    value: Optional[T]
    is_ok: bool
    msg: str = ""
    exception: Optional[Exception] = None

    def get_value(self, default: T) -> T:
        if self.is_ok:
            return self.value
        else:
            return default

    @classmethod
    def bad(cls, msg: str = "", exception: Optional[Exception] = None) -> 'ParsingResult':
        return ParsingResult(
            value=None, is_ok=False, msg=msg, exception=exception
        )

    @classmethod
    def good(cls, value: T, msg: str = "") -> 'ParsingResult':
        return ParsingResult(
            value=value, is_ok=True, msg=msg
        )



from .assumption_spec import AssumptionSpec
from .cell_extraction_spec import CellExtractionSpec
from .locator_spec import LocatorSpec
from .validator_spec import ValidatorSpec
from .excel_processor_spec import ExcelProcessorSpec



from dataclasses import dataclass
from typing import Any

from exco.exception import ParsingFailException
from exco.extractor.parser.built_in.value_parser import ValueParser


@dataclass
class IntParser(ValueParser[int]):

    def parse_value(self, v: Any) -> int:
        try:
            return int(v)  # fix this it shouldn't parse float
        except Exception as e:
            raise ParsingFailException(f'Unable to parse {v} to int') from e



from exco.extractor.validator.built_in.value_validator import ValueValidator


class IsNotBlankValidator(ValueValidator[str]):
    def validate_value(self, value: str) -> bool:
        return value == ''



from dataclasses import dataclass

from exco.extractor.validator.built_in.value_validator import ValueValidator


@dataclass
class BetweenValidator(ValueValidator[float]):
    low: float
    high: float

    def validate_value(self, v: float):
        return self.low <= v <= self.high



from dataclasses import dataclass
from typing import Optional


@dataclass
class AssumptionResult:
    is_ok: bool
    msg: str = ""
    exception: Optional[Exception] = None

    @classmethod
    def good(cls):
        return AssumptionResult(is_ok=True)

    @classmethod
    def bad(cls, msg: str = '', exception: Optional[Exception] = None):
        return AssumptionResult(is_ok=False, msg=msg, exception=exception)



import abc

from exco.extractor.actor import Actor
from exco.extractor.table_end_conditions.table_end_condition_param import TableEndConditionParam
from exco.extractor.table_end_conditions.table_end_condition_result import TableEndConditionResult


class TableEndCondition(Actor, abc.ABC):
    """Abstract class for TableEndCondition"""

    @abc.abstractmethod
    def test(self, param: TableEndConditionParam) -> TableEndConditionResult:
        """abstract method on how each implementation decides where the table should end"""
        raise NotImplementedError()



from dataclasses import dataclass, field
from typing import Dict, Any

from exco.extractor_spec.type import SpecParam
from exco.setting import k_key
from exco.util import name_params


@dataclass
class ValidatorSpec:
    """
    Validation is something to check after parsing
    Ex: if the parsed value is greater than 99
    # TODO: Maybe I should put a source here too
    """
    name: str
    params: SpecParam = field(default_factory=dict)

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> 'ValidatorSpec':
        """Construct ValidationTaskSpec from dict
        {name: greater_than, threshold: 99}

        Args:
            d (Dict[str, Any]):

        Returns:
            ValidationTaskSpec
        """
        name, params = name_params(d, exclude={k_key})
        return ValidatorSpec(name=name, params=params)



import abc
from typing import Generic, TypeVar

from exco.extractor.actor import Actor
from exco.extractor.validator.validation_result import ValidationResult

T = TypeVar('T')


class Validator(Actor, abc.ABC, Generic[T]):
    @abc.abstractmethod
    def validate(self, value: T) -> ValidationResult:
        raise NotImplementedError()



from dataclasses import dataclass, field
from typing import Dict, Any, ClassVar, Set

from exco.exception import ExcoException, ParserSpecCreationException
from exco.extractor_spec.type import SpecParam

from exco import setting as st


@dataclass
class ParserSpec:
    name: str
    params: SpecParam = field(default_factory=dict)

    allowed_keys: ClassVar[Set[str]] = {st.k_parser, st.k_params}

    @classmethod
    def from_dict(cls, d: Dict[str, Any]):
        try:
            return ParserSpec(
                name=d[st.k_parser],
                params=d.get(st.k_params, {})
            )
        except (ExcoException, KeyError) as e:
            raise ParserSpecCreationException(f'Can\'t create parser from {d}') from e



import re
from dataclasses import dataclass

from exco import CellLocation, util
from exco.extractor.locator.locating_result import LocatingResult
from exco.extractor.locator.locator import Locator
from openpyxl import Workbook
from openpyxl.worksheet.worksheet import Worksheet


@dataclass
class RightOfLocator(Locator):  # TODO: Add search scope
    label: str

    def locate(self, anchor_cell_location: CellLocation, workbook: Workbook) -> LocatingResult:
        sheet: Worksheet = workbook[anchor_cell_location.sheet_name]
        for row in sheet.iter_rows():
            for cell in row:
                if cell.value == self.label:
                    cell_loc = CellLocation(
                        sheet_name=anchor_cell_location.sheet_name,
                        coordinate=util.shift_coord(cell.coordinate, (0, 1))
                    )
                    return LocatingResult.good(cell_loc)
        return LocatingResult.bad(msg=f"Unable to find cell to the right of {self.label}")



from dataclasses import dataclass
from typing import Tuple

from exco.excel_extraction_scope import ExcelExtractionScope
from exco.util import tuple_to_coordinate
from openpyxl import Workbook
from openpyxl.utils import coordinate_to_tuple
from exco.cell_full_path import CellFullPath
from openpyxl.worksheet.worksheet import Worksheet


@dataclass(frozen=True)
class CellLocation(ExcelExtractionScope):
    """Workbook agnostic cell location."""
    sheet_name: str
    coordinate: str

    def __hash__(self):
        return hash((self.sheet_name, self.coordinate))

    def shift_row(self, offset: int) -> 'CellLocation':
        """Shift the cell by given offset in row direction

        Args:
            offset (int):

        Returns:
            CellLocation
        """
        row, col = self.row_col
        return CellLocation(sheet_name=self.sheet_name, coordinate=tuple_to_coordinate(row + offset, col))

    def shift_col(self, offset: int) -> 'CellLocation':
        """Shift the cell by given offset in column direction.

        Args:
            offset (int):

        Returns:
            CellLocation
        """
        row, col = self.row_col
        return CellLocation(sheet_name=self.sheet_name, coordinate=tuple_to_coordinate(row, col + offset))

    @property
    def short_name(self) -> str:
        """

        Returns:
            short name. Ex: SHEET1!B9
        """
        return f"{self.sheet_name}!{self.coordinate}"

    @property
    def row(self) -> int:
        """

        Returns:
            row number.
        """
        r, _ = self.row_col
        return r

    @property
    def col(self) -> int:
        """

        Returns:
            column number
        """
        _, c = self.row_col
        return c

    @property
    def row_col(self) -> Tuple[int, int]:
        """

        Returns:
            Tuple[int, int]. row, col
        """
        return coordinate_to_tuple(self.coordinate)

    def get_cell_full_path(self, wb: Workbook) -> CellFullPath:
        """ Obtain cell full path

        Args:
            wb (Workbook):

        Returns:
            CellFullPath
        """
        sheet: Worksheet = wb[self.sheet_name]
        return CellFullPath(
            workbook=wb,
            sheetname=self.sheet_name,
            sheet=sheet,
            cell=sheet[self.coordinate]
        )

    def offset_to(self, other: 'CellLocation') -> Tuple[int, int]:
        """Compute offset to another cell
            The offset returned is such that  # our + offset = other
        Args:
            other (CellLocation):

        Returns:
            Tuple[int, int]. row, col
        """
        our_r, our_c = self.row_col
        other_r, other_c = other.row_col
        return other_r - our_r, other_c - our_c



from typing import Dict, Type

from exco import ValidatorSpec
from exco.extractor.base_factory import BaseFactory
from exco.extractor.validator.built_in.between_validator import BetweenValidator
from exco.extractor.validator.built_in.is_not_blank_validator import IsNotBlankValidator
from exco.extractor.validator.validator import Validator


class ValidatorFactory(BaseFactory[Validator, ValidatorSpec]):
    def __init__(self, class_map: Dict[str, Type[Validator]]):
        super().__init__(class_map)

    @classmethod
    def suffix(cls):
        return 'Validator'

    @classmethod
    def default(cls) -> 'ValidatorFactory':
        return cls(cls.build_class_dict([
            IsNotBlankValidator,
            BetweenValidator
        ]))



from typing import Dict, Type

from exco.extractor.base_factory import BaseFactory
from exco.extractor.table_end_conditions.built_in.all_blank_table_end_condition import AllBlankTableEndCondition
from exco.extractor.table_end_conditions.built_in.max_row_table_end_condition import MaxRowTableEndCondition
from exco.extractor.table_end_conditions.table_end_condition import TableEndCondition
from exco.extractor_spec.table_extraction_spec import TableEndConditionSpec


class TableEndConditionFactory(BaseFactory[TableEndCondition, TableEndConditionSpec]):
    def __init__(self, class_map: Dict[str, Type[TableEndCondition]]):
        super().__init__(class_map)

    @classmethod
    def suffix(cls) -> str:
        """TableEndCondition class's suffix"""
        return 'TableEndCondition'

    @classmethod
    def default(cls) -> 'TableEndConditionFactory':
        """

        Returns:
            Default TableEncConditionFactory.
        """
        return cls(cls.build_class_dict([
            AllBlankTableEndCondition,
            MaxRowTableEndCondition
        ]))



from dataclasses import dataclass, field
from typing import Dict, Any, Optional

from exco import setting
from exco.util import name_params


@dataclass
class LocatorSpec:
    name: str
    params: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def default(cls):
        return LocatorSpec(name=setting.default_locator)

    @classmethod
    def from_dict(cls, d: Optional[Dict[str, Any]]) -> 'LocatorSpec':
        if d is None:
            return cls.default()
        else:
            name, params = name_params(d)
            return LocatorSpec(name=name, params=params)



from dataclasses import dataclass
from datetime import date, datetime
from typing import Any

from exco.exception import ParsingFailException
from exco.extractor.parser.built_in.value_parser import ValueParser


@dataclass
class DateParser(ValueParser[date]):

    def parse_value(self, v: Any) -> date:
        if isinstance(v, datetime):
            return v.date()
        else:
            raise ParsingFailException(f'{v} is not datetime')



from dataclasses import dataclass
from typing import Any

from exco.exception import ParsingFailException
from exco.extractor.parser.built_in.value_parser import ValueParser


@dataclass
class FloatParser(ValueParser[float]):

    def parse_value(self, v: Any) -> float:
        try:
            return float(v)
        except (TypeError, ValueError) as e:
            raise ParsingFailException(f'Fail parsing {v} to float.') from e



import abc
from typing import Type, TypeVar

from exco.cell_full_path import CellFullPath
from exco.extractor.assumption.assumption_result import AssumptionResult
from exco.extractor.actor import Actor

T = TypeVar('T')


class Assumption(Actor, abc.ABC):
    @abc.abstractmethod
    def assume(self, cfp: CellFullPath) -> AssumptionResult:
        raise NotImplementedError()



from dataclasses import dataclass
from typing import TypeVar, Dict, Any, List, Optional, Generic

import openpyxl
from exco.cell_location import CellLocation
from exco.exception import ExcoException, ExtractionTaskCreationException, TableExtractionTaskCreationException
from exco.exco_template import ExcoTemplate
from exco.extractor.assumption.assumption_factory import AssumptionFactory
from exco.extractor.cell_extraction_task import CellExtractionTaskResult, CellExtractionTask
from exco.extractor.locator.locator_factory import LocatorFactory
from exco.extractor.parser.parser_factory import ParserFactory
from exco.extractor.table_end_conditions.table_end_condition_factory import TableEndConditionFactory
from exco.extractor.table_extraction_task import TableExtractionTask, EndConditionCollection, TableExtractionTaskResult
from exco.extractor.validator.validator_factory import ValidatorFactory
from exco.extractor_spec import CellExtractionSpec, ExcelProcessorSpec
from exco.extractor_spec.table_extraction_spec import TableExtractionSpec
from openpyxl import Workbook

T = TypeVar('T')


@dataclass
class ProcessorKey:
    cell_location: CellLocation
    key: str

    def __hash__(self):
        return hash((self.cell_location, self.key))


@dataclass
class LookupResult(Generic[T]):
    cell_location: CellLocation
    result: T


@dataclass
class ExcelProcessingResult:
    cell_results: Dict[CellLocation, List[CellExtractionTaskResult]]
    table_results: Dict[CellLocation, List[TableExtractionTaskResult]]

    @property
    def is_ok(self):
        return all(cr.is_ok for crs in self.cell_results.values() for cr in crs) and \
               all(tr.is_ok for trs in self.table_results.values() for tr in trs)

    def _lookup_for_key(self, d: Dict[CellLocation, List[T]], key: str) -> Optional[LookupResult[T]]:
        for cl, results in d.items():
            for result in results:
                if key == result.key:
                    return LookupResult(cl, result)
            # TODO: Fix this O(n) search.
        return None

    def cell_result_for_key(self, key: str) -> Optional[LookupResult[CellExtractionTaskResult]]:
        return self._lookup_for_key(self.cell_results, key)

    def table_result_for_key(self, key: str) -> Optional[LookupResult[TableExtractionTaskResult]]:
        return self._lookup_for_key(self.table_results, key)

    def to_dict(self) -> Dict[str, Any]:
        ret = {}
        for results in self.cell_results.values():  # TODO: Warn of duplicate key
            for result in results:
                ret[result.key] = result.get_value(None)

        for results in self.table_results.values():  # TODO: Warn of duplicate key
            for result in results:
                ret[result.key] = result.get_value()
        return ret


@dataclass
class ExcelProcessor:
    cell_processors: Dict[CellLocation, List[CellExtractionTask]]
    table_processors: Dict[CellLocation, List[TableExtractionTask]]

    def process_workbook(self, workbook: Workbook) -> ExcelProcessingResult:
        cell_result = {}
        for loc, cets in self.cell_processors.items():
            cell_result[loc] = [cet.process(loc, workbook) for cet in cets]

        table_result = {}
        for loc, tets in self.table_processors.items():
            table_result[loc] = [tet.process(loc, workbook) for tet in tets]
        return ExcelProcessingResult(cell_results=cell_result, table_results=table_result)

    def process_excel(self, fname: str) -> ExcelProcessingResult:
        wb = openpyxl.load_workbook(fname)
        return self.process_workbook(wb)

    def __str__(self):
        tmp = []
        for cl, tasks in self.cell_processors.items():
            header = f'Location: {cl.short_name}\n'
            sep = '*\n' * 10
            task_str = sep.join([str(task) + '\n' for task in tasks])
            tmp.append(header + task_str)
        ret = ("#" * 20 + '\n').join(tmp)
        return ret


@dataclass
class ExcelProcessorFactory:
    locator_factory: LocatorFactory
    assumption_factory: AssumptionFactory
    parser_factory: ParserFactory
    validator_factory: ValidatorFactory
    table_end_condition_factory: TableEndConditionFactory

    @classmethod
    def default(cls):
        return ExcelProcessorFactory(
            locator_factory=LocatorFactory.default(),
            assumption_factory=AssumptionFactory.default(),
            parser_factory=ParserFactory.default(),
            validator_factory=ValidatorFactory.default(),
            table_end_condition_factory=TableEndConditionFactory.default()
        )

    def create_extraction_task(self, spec: CellExtractionSpec) -> CellExtractionTask:
        try:
            return CellExtractionTask(
                key=spec.key,
                locator=self.locator_factory.create_from_spec(spec=spec.locator),
                assumptions={k: self.assumption_factory.create_from_spec(sp) for k, sp in
                             spec.assumptions.items()},
                parser=self.parser_factory.create_from_spec(spec=spec.parser),
                validators={k: self.validator_factory.create_from_spec(sp) for k, sp in
                            spec.validations.items()}
            )
        except ExcoException as e:
            raise ExtractionTaskCreationException(
                f'Unable to create ExtractionTask for {spec.key} cf {spec.source.describe()}') from e

    def create_table_extraction_task(self, spec: TableExtractionSpec) -> TableExtractionTask:
        columns = {k: self.create_extraction_task(CellExtractionSpec(v))
                   for k, v in spec.columns.items()}
        try:
            return TableExtractionTask(
                key=spec.key,
                locator=self.locator_factory.create_from_spec(spec=spec.locator),
                columns=columns,
                end_condition=EndConditionCollection.from_spec(spec.end_conditions,
                                                               factory=self.table_end_condition_factory),
                item_direction=spec.item_direction
            )
        except ExcoException as e:
            raise TableExtractionTaskCreationException(
                f'Unable to create TableExtractionTask for {spec.key} cf\n {spec.source.describe()}') from e

    def create_from_spec(self, spec: ExcelProcessorSpec) -> ExcelProcessor:
        cell_tasks = {}
        for cl, specs in spec.cell_specs.items():
            cell_tasks[cl] = [self.create_extraction_task(spec) for spec in specs]

        table_tasks = {}
        for cl, specs in spec.table_specs.items():
            table_tasks[cl] = [self.create_table_extraction_task(spec) for spec in specs]

        return ExcelProcessor(cell_processors=cell_tasks, table_processors=table_tasks)

    def create_from_template_excel(self, fname: str) -> ExcelProcessor:
        workbook = openpyxl.load_workbook(fname)
        return self.create_from_template_workbook(workbook)

    def create_from_template_workbook(self, workbook: Workbook) -> ExcelProcessor:
        template = ExcoTemplate.from_workbook(workbook)
        spec = template.to_excel_extractor_spec()
        return self.create_from_spec(spec)



from typing import Dict, Type

from exco import LocatorSpec
from exco.extractor.base_factory import BaseFactory
from exco.extractor.locator.built_in.at_comment_cell_locator import AtCommentCellLocator
from exco.extractor.locator.built_in.right_of_locator import RightOfLocator
from exco.extractor.locator.built_in.right_of_regex_locator import RightOfRegexLocator
from exco.extractor.locator.locator import Locator


class LocatorFactory(BaseFactory[Locator, LocatorSpec]):
    def __init__(self, class_map: Dict[str, Type[Locator]]):
        super().__init__(class_map)

    @classmethod
    def suffix(cls):
        return 'Locator'

    @classmethod
    def default(cls) -> 'LocatorFactory':
        return cls(cls.build_class_dict([
            AtCommentCellLocator,
            RightOfLocator,
            RightOfRegexLocator
        ]))



from dataclasses import dataclass
from typing import Any

from exco.exception import ParsingFailException
from exco.extractor.parser.built_in.value_parser import ValueParser


@dataclass
class StringParser(ValueParser[str]):

    def parse_value(self, v: Any) -> str:
        return '' if v is None else str(v)



import abc
from typing import Any, TypeVar, Type
from typing import Dict

T = TypeVar('T')


class Actor(abc.ABC):
    @classmethod
    def create(cls: Type[T], params: Dict[str, Any]) -> T:
        return cls(**params)



import abc
from typing import TypeVar

from exco.cell_full_path import CellFullPath
from openpyxl import Workbook

T = TypeVar('T')


class ExcelExtractionScope(abc.ABC):

    @abc.abstractmethod
    def get_cell_full_path(self, wb: Workbook) -> CellFullPath:
        raise NotImplementedError()



from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Any, List, Optional

from exco import setting as st
from exco.extractor_spec.apv_spec import APVSpec
from exco.extractor_spec.locator_spec import LocatorSpec
from exco.extractor_spec.spec_source import UnknownSource, SpecSource
from exco.extractor_spec.type import SpecParam
from exco.util import name_params


class TableItemDirection(Enum):
    DOWNWARD = 'downward'
    RIGHTWARD = 'rightward'

    @classmethod
    def default(cls) -> 'TableItemDirection':
        return TableItemDirection.DOWNWARD

    @classmethod
    def from_value(cls, v: Optional[str]) -> 'TableItemDirection':
        if v is None:
            return cls.default()
        else:
            return TableItemDirection(v)


Offset = int


@dataclass
class TableEndConditionSpec:
    name: str
    params: SpecParam = field(default_factory=dict)

    @classmethod
    def default_conditions(cls) -> List['TableEndConditionSpec']:
        return [TableEndConditionSpec(name='all_blank', params={})]

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> 'TableEndConditionSpec':
        name, params = name_params(d)
        return TableEndConditionSpec(name=name, params=params)


@dataclass
class ColumnSpecDict:
    offset: int
    dict: Dict[str, Any]
    source: SpecSource


@dataclass
class TableExtractionSpec:
    key: str
    locator: LocatorSpec  # for corner
    columns: Dict[Offset, APVSpec]  # offset -> APVSpec
    end_conditions: List[TableEndConditionSpec] = field(default_factory=TableEndConditionSpec.default_conditions)
    item_direction: TableItemDirection = TableItemDirection.DOWNWARD
    source: SpecSource = field(default_factory=UnknownSource)

    @classmethod
    def from_dict(cls, d: Dict[str, Any], source: SpecSource = None):
        """Construct form dict

        Args:
            d ():
            source ():

        Returns:

        """
        source = source if source is not None else UnknownSource()
        col_dicts = [ColumnSpecDict(offset, col_spec, UnknownSource()) for offset, col_spec in d[st.k_columns].items()]
        return cls.from_table_and_column_dict(
            d, source, col_dicts
        )

    @classmethod
    def from_table_and_column_dict(cls, d: Dict[str, Any],
                                   source: SpecSource,
                                   column_dicts: List[ColumnSpecDict]) -> 'TableExtractionSpec':
        """
        Example:

        key: some_key
        locator: {LocatorSpec} # optional
        columns:
            - {offset: offset, spec: {AVPSpec}}
        end_conditions: # optional
            - {EndConditionSpec}
        item_direction: downward # optional

        Args:
            d (Dict[str, Any]):
            source (SpecSource):
            column_dicts (List[ColumnSpecDict])

        Returns:
            TableExtractionSpec
        """
        columns = {
            col.offset: APVSpec.from_dict(d=col.dict, source=col.source)
            for col in column_dicts
        }
        if st.k_end_conditions in d:
            end_conditions = [TableEndConditionSpec.from_dict(v) for v in d.get(st.k_end_conditions, [])]
        else:
            end_conditions = TableEndConditionSpec.default_conditions()  # should i have this default??
        return TableExtractionSpec(
            key=d[st.k_key],
            locator=LocatorSpec.from_dict(d.get(st.k_locator, None)),
            columns=columns,
            end_conditions=end_conditions,
            item_direction=TableItemDirection.from_value(d.get(st.k_item_direction, None)),
            source=source
        )



import abc
from typing import Generic, TypeVar

from exco.extractor.validator.validator import Validator
from exco.extractor.validator.validation_result import ValidationResult

T = TypeVar('T')


class ValueValidator(Generic[T], Validator[T], abc.ABC):
    @abc.abstractmethod
    def validate_value(self, value: T) -> bool:
        raise NotImplementedError()

    def validate(self, value: T) -> ValidationResult:
        if self.validate_value(value):
            return ValidationResult.good()
        else:
            return ValidationResult.bad(msg=f"Fail {value} fail validation of {self}")



from dataclasses import dataclass, field
from typing import Dict, Any, ClassVar, Set

from exco import util, exception
from exco.extractor_spec.validator_spec import ValidatorSpec
from exco.extractor_spec.apv_spec import APVSpec
from exco.extractor_spec.locator_spec import LocatorSpec
from exco.extractor_spec.parser_spec import ParserSpec
from exco.extractor_spec.spec_source import SpecSource, UnknownSource
from exco import setting as st
from exco.extractor_spec.assumption_spec import AssumptionSpec


@dataclass
class CellExtractionSpec:
    apv: APVSpec
    locator: LocatorSpec = field(default_factory=LocatorSpec.default)
    source: SpecSource = field(default_factory=UnknownSource)

    consumed_keys: ClassVar[Set[str]] = {st.k_locator}
    allowed_keys: ClassVar[Set[str]] = consumed_keys | APVSpec.allowed_keys

    @property
    def key(self) -> str:
        return self.apv.key

    @property
    def parser(self) -> ParserSpec:
        return self.apv.parser

    @property
    def validations(self) -> Dict[str, ValidatorSpec]:
        return self.apv.validations

    @property
    def assumptions(self) -> Dict[str, AssumptionSpec]:
        return self.apv.assumptions

    @classmethod
    def from_dict(cls, d: Dict[str, Any], source: SpecSource = None) -> 'CellExtractionSpec':
        source = source if source is not None else UnknownSource()
        extra_keys = util.extra_keys(d, cls.allowed_keys)
        if extra_keys:
            raise exception.ExcoBlockContainsExtraKey(f'{extra_keys}\n'
                                                      f'allowed_keys are {cls.allowed_keys}')
        return CellExtractionSpec(
            locator=LocatorSpec.from_dict(d.get(st.k_locator, None)),
            apv=APVSpec.from_dict(d, source),
            source=source
        )



import abc
from typing import Any, TypeVar, Type, Dict

from exco.cell_full_path import CellFullPath
from exco.exception import ParsingFailException
from exco.extractor.parser.parser import Parser
from exco.extractor.parser.parsing_result import ParsingResult

T = TypeVar('T')


class ValueParser(Parser[T], abc.ABC):

    @abc.abstractmethod
    def parse_value(self, v: Any) -> T:
        raise NotImplementedError()

    def parse(self, cfp: CellFullPath) -> ParsingResult[T]:
        try:
            return ParsingResult.good(self.parse_value(cfp.cell.value))
        except ParsingFailException as e:
            return ParsingResult.bad(f'Parsing {cfp.cell.coordinate} fail.\n'
                                     f'value => "{cfp.cell.value}"\n'
                                     f'{e.msg}',
                                     exception=e)



import abc
from typing import Generic, TypeVar

from exco.cell_full_path import CellFullPath
from exco.extractor.actor import Actor

from exco.extractor.parser.parsing_result import ParsingResult

PARSERCLASS_SUFFIX = 'Parser'
T = TypeVar('T')


class Parser(Actor, abc.ABC, Generic[T]):
    @abc.abstractmethod
    def parse(self, cfp: CellFullPath) -> ParsingResult[T]:
        raise NotImplementedError()



import itertools
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List

import openpyxl as opx
from exco import util
from exco.cell_location import CellLocation
from exco.exception import ExcoException, BadTemplateException, CommentWithNoExcoBlockWarning, TableKeyNotFound, \
    TableHasNoColumn, MissingTableBlock
from exco.exco_template.exco_block import ExcoBlock, ExcoBlockCollection
from exco.extractor_spec.cell_extraction_spec import CellExtractionSpec
from exco.extractor_spec.excel_processor_spec import ExcelProcessorSpec
from exco.extractor_spec.spec_source import SpecSource
from exco.extractor_spec.table_extraction_spec import TableExtractionSpec, ColumnSpecDict
from openpyxl.cell.read_only import EmptyCell
import warnings


@dataclass
class ExcoBlockWithLocation(SpecSource):
    cell_location: CellLocation
    exco_block: ExcoBlock

    @property
    def key(self) -> str:
        """

        Returns:
            Block's key
        """
        return self.exco_block.key

    def describe(self) -> str:
        """

        Returns:
            Block description.
        """
        return self.cell_location.sheet_name + '\n' + self.exco_block.raw

    @classmethod
    def simple(cls, raw: str,
               sheet: str = 'SHEET1',
               coord: str = 'A1',
               start_line: int = 1,
               end_line: int = 1):
        return ExcoBlockWithLocation(
            cell_location=CellLocation(sheet, coord),
            exco_block=ExcoBlock(
                start_line=start_line,
                end_line=end_line,
                raw=raw
            )
        )


@dataclass
class ExcoTemplate:
    table_blocks: List[ExcoBlockWithLocation]
    column_blocks: List[ExcoBlockWithLocation]
    cell_blocks: List[ExcoBlockWithLocation]

    @classmethod
    def empty(cls) -> 'ExcoTemplate':
        """Create Empty Template

        Returns:
            ExcoTemplate
        """
        return ExcoTemplate(table_blocks=[], column_blocks=[], cell_blocks=[])

    def add_to_block_collections(self, cell_location: CellLocation, block_collection: ExcoBlockCollection) -> None:
        """Add block collection as if anchor cell is at cell_location.

        Args:
            cell_location (CellLocation):
            block_collection (ExcoBlockCollection):

        Returns:
            None.
        """

        for tb in block_collection.table_blocks:
            self.table_blocks.append(ExcoBlockWithLocation(cell_location, tb))
        for cell_block in block_collection.cell_blocks:
            self.cell_blocks.append(ExcoBlockWithLocation(cell_location, cell_block))
        for col_block in block_collection.column_blocks:
            self.column_blocks.append(ExcoBlockWithLocation(cell_location, col_block))

    def cell_locations(self) -> List[CellLocation]:
        """

        Returns:
            List of unique cell locations
        """
        return util.unique(x.cell_location for x in itertools.chain(
            self.table_blocks,
            self.column_blocks,
            self.cell_blocks
        ))

    def n_cell(self) -> int:
        """

        Returns:
            Total number of cells with comments. Some comment may have no exco block.
        """
        return len(self.cell_locations())

    def n_exco_blocks(self) -> int:
        """

        Returns:
            Total number of ExcoBlocks
        """
        return len(self.table_blocks) + len(self.column_blocks) + len(self.cell_blocks)

    @classmethod
    def from_workbook(cls, workbook: opx.Workbook) -> 'ExcoTemplate':
        """Construct from Excel workbook

        Args:
            workbook (workbook):

        Returns:
            ExcoTemplate
        """
        wb = workbook
        ret = ExcoTemplate.empty()
        for cfp in util.iterate_cells_in_workbook(wb):
            if not isinstance(cfp.cell, EmptyCell) and cfp.cell.comment is not None:
                cell_loc = cfp.to_cell_location()
                try:
                    ebc = ExcoBlockCollection.from_string(cfp.cell.comment.text)
                    if ebc.n_total_blocks() == 0:
                        warnings.warn(f"{cell_loc.short_name} has comment but no exco block.",
                                      CommentWithNoExcoBlockWarning)
                    ret.add_to_block_collections(cell_loc, ebc)
                except ExcoException as e:  # throw error with cell info
                    raise BadTemplateException(  # Todo: maybe be all these should be warning
                        f'Bad Template at sheet: {cell_loc.sheet_name}, cell: {cell_loc.coordinate}') from e
        return ret

    @classmethod
    def from_excel(cls, fname: str) -> 'ExcoTemplate':
        """Construct from excel file.

        Args:
            fname (str): file name.

        Returns:
            ExcoTemplate
        """
        return cls.from_workbook(workbook=opx.load_workbook(fname))

    def column_block_dict_by_table_key(self) -> Dict[str, List[ExcoBlockWithLocation]]:
        """ Compute column block dicationary grouped by table key

        Returns:
            Dict[str, List[ExcoBlockWithLocation]]. TableKey -> [Column's ExcoBlock]
        """

        def get_table_key(block: ExcoBlockWithLocation):
            try:
                return block.exco_block.table_key()
            except ExcoException as e:
                raise TableKeyNotFound(f'{block.cell_location.short_name}') from e

        return util.group_by(get_table_key, self.column_blocks)

    def build_table_specs(self) -> Dict[CellLocation, List[TableExtractionSpec]]:
        """build table spec

        Returns:
            Dict[CellLocation, List[TableExtractionSpec]]. Anchor -> List[TableExtractionSpec]
        """
        group_columns = self.column_block_dict_by_table_key()
        used_keys = []
        ret: Dict[CellLocation, List[TableExtractionSpec]] = defaultdict(list)
        for tb in self.table_blocks:
            table_key = tb.key
            table_loc = tb.cell_location
            used_keys.append(table_key)
            try:
                column_blocks = group_columns[table_key]
            except LookupError as e:
                raise TableHasNoColumn(table_key) from e
            # TODO: Support other orientation by allowing tuple
            columns_dicts = []
            for cb in column_blocks:
                offset = tb.cell_location.offset_to(cb.cell_location)[1]
                columns_dicts.append(ColumnSpecDict(
                    offset=offset, dict=cb.exco_block.to_dict(), source=cb
                ))
            table_spec = TableExtractionSpec.from_table_and_column_dict(
                d=tb.exco_block.to_dict(),
                source=tb,
                column_dicts=columns_dicts
            )
            ret[table_loc].append(table_spec)

        if len(group_columns) != len(used_keys):
            missing_key = [gc for gc in group_columns.keys() if gc not in used_keys]
            raise MissingTableBlock(f"{missing_key} has no matching table block.")
        return dict(ret)

    def build_cell_specs(self) -> Dict[CellLocation, List[CellExtractionSpec]]:
        """build cell specs

        Returns:
             Dict[CellLocation, List[CellExtractionSpec]]. Anchor to CellExtractionSpec.
        """

        ret: Dict[CellLocation, List[CellExtractionSpec]] = defaultdict(list)

        for cb in self.cell_blocks:
            cl = cb.cell_location
            try:
                ret[cl].append(CellExtractionSpec.from_dict(cb.exco_block.to_dict(), cb))
            except ExcoException as e:
                raise BadTemplateException(cl.short_name) from e
        return dict(ret)

    def to_excel_extractor_spec(self) -> ExcelProcessorSpec:
        """compute excel extractor spec

        Returns:
            ExcelProcessorSpec
        """
        return ExcelProcessorSpec(
            cell_specs=self.build_cell_specs(),
            table_specs=self.build_table_specs())



from typing import Dict, Type

from exco.extractor_spec.parser_spec import ParserSpec
from exco.extractor.base_factory import BaseFactory
from exco.extractor.parser.built_in.date_parser import DateParser
from exco.extractor.parser.parser import Parser
from exco.extractor.parser.built_in.string_parser import StringParser
from exco.extractor.parser.built_in.int_parser import IntParser
from exco.extractor.parser.built_in.float_parser import FloatParser


class ParserFactory(BaseFactory[Parser, ParserSpec]):
    def __init__(self, class_map: Dict[str, Type[Parser]]):
        super().__init__(class_map)

    @classmethod
    def suffix(cls):
        return 'Parser'

    @classmethod
    def default(cls) -> 'ParserFactory':
        return cls(cls.build_class_dict([
            IntParser,
            StringParser,
            FloatParser,
            DateParser
        ]))



import abc


class SpecSource(abc.ABC):
    @abc.abstractmethod
    def describe(self) -> str:
        """

        Returns:
            str to print incase there is an error constructing extractor for tracing back
        """
        raise NotImplementedError()

class UnknownSource(SpecSource):

    def describe(self) -> str:
        return 'Unknown Source'




from exco.__version__ import version
from exco import util
from exco import exception
from exco.cell_location import CellLocation
from exco.exco_template import ExcoTemplate, ExcoBlock
from exco.extractor_spec import AssumptionSpec, \
    ValidatorSpec, \
    CellExtractionSpec, \
    LocatorSpec, \
    ExcelProcessorSpec

from exco.extractor import ExcelProcessorFactory, ExcelProcessor
from exco.shortcut import from_excel



from dataclasses import dataclass
from typing import Dict

from exco.cell_full_path import CellFullPath


@dataclass
class TableEndConditionParam:
    """Parameter to send to End Condition 's test function"""
    row_count: int
    """row count including this row"""
    cfps: Dict[str, CellFullPath]
    """dictionary from column key to cell full path"""



from dataclasses import dataclass

from openpyxl import Workbook

from exco import CellLocation
from exco.extractor.locator.locating_result import LocatingResult
from exco.extractor.locator.locator import Locator


@dataclass
class AtCommentCellLocator(Locator):

    def locate(self, anchor_cell_location: CellLocation, workbook: Workbook) -> LocatingResult:
        return LocatingResult.good(anchor_cell_location)



from os.path import join, dirname

from setuptools import setup, find_packages


def get_version():
    fname = join(dirname(__file__), "src/exco/__version__.py")
    with open(fname) as f:
        ldict = {}
        code = f.read()
        exec(code, globals(), ldict)  # version defined here
        return ldict['version']


setup(name='exco',
      version=get_version(),
      description='Excel Comment Orm',
      long_description=open('README.md').read().strip(),
      long_description_content_type='text/markdown',
      author='Piti Ongmongkolkul',
      author_email='piti118@gmail.com',
      url='https://github.com/thegangtechnology/exco',
      package_dir={'': 'src'},
      packages=find_packages('src'),
      py_modules=['exco'],
      entry_points={
          'console_scripts': ['exco_watch=exco.exco_watch:ExcoWatch.main'],
      },
      install_requires=[
          'openpyxl',
          'pyyaml',
          'stringcase',
          'watchdog'
      ],
      extras_require={
          'dev': [
              'ipykernel',
              'mypy',
              'autopep8',
              'pytest',
              'pytest-cov',
              'wheel'
          ],
          'test': [
              'pytest',
              'pytest-cov'
          ]
      },
      license='Private',
      zip_safe=False,
      keywords='',
      classifiers=[''])



from .excel_processor import ExcelProcessorFactory, ExcelProcessor
from .parser.parser import Parser
from .locator.locator import Locator
from .validator.validator import Validator
from .assumption.assumption import Assumption



from typing import Dict, Type

from exco import AssumptionSpec
from exco.extractor.assumption.assumption import Assumption
from exco.extractor.assumption.built_in.left_cell_match_assumption import LeftCellMatchAssumption
from exco.extractor.base_factory import BaseFactory


class AssumptionFactory(BaseFactory[Assumption, AssumptionSpec]):
    def __init__(self, class_map: Dict[str, Type[Assumption]]):
        super().__init__(class_map)

    @classmethod
    def suffix(cls):
        return 'Assumption'

    @classmethod
    def default(cls) -> 'AssumptionFactory':
        return cls(cls.build_class_dict([
            LeftCellMatchAssumption
        ]))



from dataclasses import dataclass

from openpyxl import Workbook
from openpyxl.cell import Cell
from openpyxl.cell.read_only import EmptyCell
from openpyxl.worksheet.worksheet import Worksheet


@dataclass
class CellFullPath:
    workbook: Workbook
    sheetname: str  # todo: make sheetname a property from sheet.title
    sheet: Worksheet
    cell: Cell

    def is_blank(self):
        """

        Returns:
            True if cell is blank.
        """
        return isinstance(self.cell, EmptyCell) or self.cell.value is None

    def shift(self, row: int = 0, col: int = 0) -> 'CellFullPath':
        """Shift cell full path by row and column

        Args:
            row (int):
            col (int):

        Returns:
            CellFullPath
        """
        r, c = self.cell.row, self.cell.column
        cell = self.sheet.cell(row=r + row, column=c + col)
        return CellFullPath(
            workbook=self.workbook,
            sheetname=self.sheetname,
            sheet=self.sheet,
            cell=cell
        )

    def to_cell_location(self) -> 'CellLocation':
        """Convert to workbook agnostic cell location.
        Returns:
            CellLocation
        """
        from exco.cell_location import CellLocation
        return CellLocation(sheet_name=self.sheetname, coordinate=self.cell.coordinate)



import abc
from typing import TypeVar, Generic, Dict, Type, List

from exco import util
from exco.exception import ActorCreationFailException
from exco.extractor.actor import Actor

ActorType = TypeVar('ActorType', bound=Actor)
SpecType = TypeVar('SpecType')
T = TypeVar('T')


class BaseFactory(abc.ABC, Generic[ActorType, SpecType]):
    def __init__(self, class_map: Dict[str, Type[ActorType]]):
        self.class_map = class_map

    @classmethod
    @abc.abstractmethod
    def suffix(cls):
        """class suffix for this factory

        Returns:

        """
        raise NotImplementedError()

    @classmethod
    def key_for_class(cls, clz: Type[ActorType]) -> str:
        """Compute Key for class. By default it's suffix removed snake-cased of the class name
        Ex: AllBlankTableEndCondition -> all_blank (suffix is specified in suffix())
        Args:
            clz (Type[ActorType]):

        Returns:
            str
        """
        return util.default_key(clz, suffix=cls.suffix())

    @classmethod
    def build_class_dict(cls, classes: List[ActorType]) -> Dict[str, Type[ActorType]]:
        """Build class dictionary.
        The key is by default the class name with the suffix removed and then snake case.
        Ex: AllBlankTableEndCondition -> all_blank (suffix is specified in suffix())

        Args:
            classes (List[ActorType]):

        Returns:
            Dict[str, Type[ActorType]]
        """
        return {cls.key_for_class(clz): clz for clz in classes}

    def register(self, clz: Type[ActorType]):
        """Register a single class
        Args:
            clz ():

        Returns:
            None
        """
        key = self.key_for_class(clz)
        self.class_map[key] = clz

    def register_all(self, classes: List[Type[ActorType]]):
        """Register All actor classes
        Args:
            classes(List[Type[ActorType]]) :
        Returns:
            None
        """
        for c in classes:
            self.register(c)

    def available_keys(self) -> List[str]:
        """

        Returns:
            List of all available keys.
        """
        return list(self.class_map.keys())

    def create_from_spec(self, spec: SpecType) -> ActorType:
        """Create Actor from the given spec

        Args:
            spec (SpecType):

        Returns:
            ActorType
        """
        suffix = self.suffix()
        if spec.name not in self.class_map:
            raise ActorCreationFailException(f'Cannot find {suffix} for name {spec.name}\n'
                                             f'available keys are {list(self.class_map.keys())}')
        clz = self.class_map[spec.name]
        try:
            return clz.create(params=spec.params)
        except Exception as e:
            raise ActorCreationFailException(
                f'Unable to construct {suffix} from spec {spec}') from e



import argparse
import time
from datetime import datetime, timedelta
from os.path import basename, dirname
from pprint import pprint
from traceback import print_exc

from watchdog.events import PatternMatchingEventHandler

import exco
from exco.exception import ExcoException


class ExcoWatchHandler(PatternMatchingEventHandler):
    def __init__(self, path: str):
        pattern = f'*/{basename(path)}'
        super().__init__(patterns=[pattern])
        self.path = path
        self.last_modified = datetime.now()

    def run_exco(self):
        try:
            processor = exco.from_excel(self.path)
            result = processor.process_excel(self.path)
            pprint(result.to_dict())
        except (ExcoException, TypeError, LookupError, ValueError, AttributeError):
            print_exc()
        finally:
            current_time = datetime.now().strftime("%H:%M:%S")
            print(f'Latest Result At: {current_time}')

    def on_modified(self, event):
        if datetime.now() - self.last_modified < timedelta(seconds=1):
            return False
        else:
            self.last_modified = datetime.now()

        self.run_exco()
        return True


class ExcoWatch:

    @classmethod
    def main(cls):

        parser = argparse.ArgumentParser(description='Continuously Watch Template and Output.')
        parser.add_argument('path', type=str,
                            help='Path to template file')

        args = parser.parse_args()
        cls.run(args.path)

    @classmethod
    def run(cls, path: str):
        """Continuously Watch and print template result to std out

            path: str
        """

        from watchdog.observers import Observer
        event_handler = ExcoWatchHandler(path=path)
        observer = Observer()
        observer.schedule(event_handler, dirname(path), recursive=False)
        event_handler.run_exco()
        observer.start()

        try:
            while True:
                time.sleep(1)
        except InterruptedError:
            print('Quitting...')
        finally:
            observer.stop()
            observer.join()



from dataclasses import dataclass

from exco.extractor.table_end_conditions.table_end_condition_param import TableEndConditionParam
from exco.extractor.table_end_conditions.table_end_condition_result import TableEndConditionResult
from exco.extractor.table_end_conditions.table_end_condition import TableEndCondition


@dataclass
class MaxRowTableEndCondition(TableEndCondition):
    """
    End Condition when maximum number of row is reached.
    """
    n: int
    inclusive: bool = True

    def test(self, param: TableEndConditionParam) -> TableEndConditionResult:
        """Return True if the the row number is greater than or equal to
        self.n.
        Args:
            param (TableEndConditionParam):

        Returns:
            TableEndConditionResult
        """
        max_row_reached = param.row_count >= self.n
        return TableEndConditionResult.good(
            should_terminate=max_row_reached,
            is_inclusive=self.inclusive
        )



class ExcoException(RuntimeError):
    def __init__(self, msg: str = ''):
        super().__init__(msg)
        self.msg = msg


class ExcoWarning(Warning):
    pass


class TooManyBeginException(ExcoException):
    pass


class TooManyEndException(ExcoException):
    pass


class ExpectEndException(ExcoException):
    pass


class BadTemplateException(ExcoException):
    pass


class CommentWithNoExcoBlockWarning(ExcoWarning):
    pass


class ExcoBlockContainsExtraKey(ExcoException):
    pass


class ActorCreationFailException(ExcoException):
    pass


class ParserCreationFailException(ExcoException):
    pass


class ParserSpecCreationException(ExcoException):
    pass


class ParsingFailException(ExcoException):
    pass


class ExtractionTaskCreationException(ExcoException):
    pass


class ExcelProcessorCreationException(ExcoException):
    pass


class TooManyRowRead(ExcoException):
    pass


class TableKeyNotFound(ExcoException):
    pass


class TableHasNoColumn(ExcoException):
    pass


class MissingTableBlock(ExcoException):
    pass


class TableExtractionTaskCreationException(ExcoException):
    pass


class NoEndConditonError(ExcoException):
    pass


class YamlParseError(ExcoException):
    pass



from dataclasses import dataclass
from typing import Optional


@dataclass
class ValidationResult:
    is_ok: bool
    msg: str = ""
    exception: Optional[Exception] = None

    @classmethod
    def good(cls) -> 'ValidationResult':
        return ValidationResult(is_ok=True)

    @classmethod
    def bad(cls, msg: str = "", exception: Optional[Exception] = None):
        return ValidationResult(
            is_ok=False,
            msg=msg,
            exception=exception
        )



from dataclasses import dataclass
from typing import Dict, List

from exco.cell_location import CellLocation
from exco import util
from exco.extractor_spec.cell_extraction_spec import CellExtractionSpec
from exco.extractor_spec.table_extraction_spec import TableExtractionSpec
from itertools import chain


@dataclass
class ExcelProcessorSpec:
    # TODO: don't group by cell loc.
    cell_specs: Dict[CellLocation, List[CellExtractionSpec]]  # anchor -> specs
    table_specs: Dict[CellLocation, List[TableExtractionSpec]]  # anchor -> specs

    def n_total_spec(self) -> int:
        return self.n_table_spec() + self.n_cell_spec()

    def n_total_location(self) -> int:
        return len(set(self.cell_specs.keys()) | set(self.table_specs.keys()))

    def n_table_spec(self) -> int:
        return util.flattened_len(self.table_specs.values())

    def n_cell_spec(self) -> int:
        return util.flattened_len(self.cell_specs.values())

    def n_cell_location(self) -> int:
        return len(self.cell_specs)

    def n_table_location(self) -> int:
        return len(self.table_specs)

    def is_keys_unique(self) -> bool:
        it = chain(
            util.flatten(self.cell_specs.values()),
            util.flatten(self.table_specs.values()))
        return util.is_unique(
            spec.key for spec in it
        )



import textwrap
from collections import defaultdict
from typing import TypeVar, Iterable, Any, Dict, List, Tuple, Type, Set, Optional, Generator, Sequence, Callable

import openpyxl
from exco.cell_full_path import CellFullPath
from openpyxl import Workbook
from openpyxl.cell import Cell

from openpyxl.utils import get_column_letter
import stringcase
from exco import setting as st
import itertools

from openpyxl.worksheet.worksheet import Worksheet

T = TypeVar('T')


def long_string(s: str) -> str:
    """left strip and dedent the string

    Args:
        s (str):

    Returns:
        left string and dedent the string
    """
    return textwrap.dedent(s).lstrip()


def is_unique(items: Iterable[Any]) -> bool:
    s = set()
    for item in items:
        if item in s:
            return False
        s.add(item)
    return True


def tuple_to_coordinate(row, col) -> str:
    return get_column_letter(col) + str(row)


def coordinate_to_tuple(coord: str) -> Tuple[int, int]:
    return openpyxl.utils.coordinate_to_tuple(coord)


def shift_coord(coord: str, shift: Tuple[int, int]):
    row, col = coordinate_to_tuple(coord)
    sr, sc = shift
    return tuple_to_coordinate(row + sr, col + sc)


def remove_suffix(s: str, suffix: str):  # polyfill for python<3.9
    """Remove suffix(if exists) from s

    Args:
        s (str):
        suffix (str): suffix to remove.

    Returns:
        suffix removed s.
    """
    if s.endswith(suffix):
        return s[:-len(suffix)]
    else:
        return s


def default_key(clz: Type[Any], suffix):
    """Return snake case with suffix removed. This is used as default key for class.

    Args:
        s (str):
        suffix (str):

    Returns:
        default key name: Ex: IntParser -> int

    """
    return stringcase.snakecase(remove_suffix(clz.__name__, suffix))


def extra_keys(d: Dict[str, Any], allowed=Set[str]) -> List[str]:
    return [k for k in d.keys() if k not in allowed]


def name_params(d: Dict[str, Any], exclude: Optional[Set[str]] = None) -> Tuple[str, Dict[str, Any]]:
    exclude = set() if exclude is None else exclude
    name = d[st.k_name]
    params = {k: v for k, v in d.items() if k != st.k_name and k not in exclude}
    return name, params


def flatten(lol: Iterable[Iterable[T]]) -> Iterable[T]:
    return itertools.chain.from_iterable(lol)


def flattened_len(it: Iterable[Iterable]):
    """Find the len of flatten list
    """
    return sum(1 for _ in flatten(it))


def iterate_cells_in_worksheet(sheet: Worksheet) -> Generator[Cell, None, None]:
    for row in sheet.iter_rows():
        for cell in row:
            yield cell


def iterate_cells_in_workbook(workbook: Workbook) -> Generator[CellFullPath, None, None]:
    for sheetname in workbook.sheetnames:
        sheet: Worksheet = workbook[sheetname]
        for cell in iterate_cells_in_worksheet(sheet):
            yield CellFullPath(
                workbook=workbook,
                sheetname=sheetname,
                sheet=sheet,
                cell=cell)


def unique(xs: Iterable[T]) -> List[T]:
    return list(set(xs))


T2 = TypeVar('T2')


def group_by(f: Callable[[T], T2], xs: List[T]) -> Dict[T2, List[T]]:
    ret = defaultdict(list)
    for x in xs:
        key = f(x)
        ret[key].append(x)
    return dict(ret)
